/*
 * Copyright (C) 2007 - John Anderson, Neil Kettle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#ifdef DEBUG
#include <assert.h>
#endif

#include <arpa/inet.h>
#include <ifaddrs.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "common.h"
#include "file.h"
#include "net.h"
#include "obuf.h"

#include "synscan.h"

#ifndef IFF_UP
#define IFF_UP 1
#endif

sock_t
net_sock_init (void)
{
  sock_t sock;

  sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
  if (sock == -1)
    return (-1);

#ifdef MACOS
  {
    int true = 1;
    setsockopt (sock, IPPROTO_IP, IP_HDRINCL, &true, sizeof true);
  }
#endif

  return (sock);
}

static u_int16_t
net_ip_sum (const u_int16_t * ptr, int len)
{
  register int sum = 0;

  while (len > 1)
    {
      sum += *ptr++;
      len -= 2;
    }

  if (len == 1)
    sum += *((unsigned char *) ptr);

  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return (~sum);
}

static void
net_ipv4_send_segment (struct net_pkt_t *pkt, struct ip_hdr_t *ip_hdr,
                       struct tcp_hdr_t *tcp_hdr, struct tcp_phdr_t *tcp_phdr)
{
  void *pbuf;
  struct sockaddr_in sin;

  pbuf = tcp_phdr;
  memcpy (pbuf + sizeof *tcp_phdr, tcp_hdr, sizeof *tcp_hdr);
  tcp_hdr->tcp_sum =
    net_ip_sum ((u_int16_t *) pbuf, sizeof *tcp_phdr + sizeof *tcp_hdr);

  pbuf = ip_hdr;
  ip_hdr->ip_sum = net_ip_sum ((u_int16_t *) pbuf, sizeof *ip_hdr);

  sin.sin_family = AF_INET;
  sin.sin_port = tcp_hdr->tcp_dport;
  sin.sin_addr = ip_hdr->ip_dst_t.ip_dst;

#ifdef DEBUG
  printf ("%04x %04x\n", ip_hdr->ip_sum, tcp_hdr->tcp_sum);
#endif

  sendto (pkt->rsock, pbuf, sizeof *ip_hdr + sizeof *tcp_hdr, 0,
          (struct sockaddr *) &sin, sizeof sin);
}

void
net_ipv4_send (struct net_pkt_t *pkt, u_int32_t srcip, u_int32_t dstip,
               u_int16_t sport, u_int16_t dport)
{
  struct ip_hdr_t *ip_hdr;
  struct tcp_hdr_t *tcp_hdr;
  struct tcp_phdr_t *tcp_phdr;

  ip_hdr = pkt->ip_hdr;
  tcp_hdr = pkt->tcp_hdr;
  tcp_phdr = pkt->tcp_phdr;

  /* update variable header values */
  ip_hdr->ip_sum = 0;
  ip_hdr->ip_src_t.ip_src_int = srcip;
  ip_hdr->ip_dst_t.ip_dst_int = dstip;

  tcp_hdr->tcp_sum = 0;
  tcp_hdr->tcp_sport = htons (sport);
  tcp_hdr->tcp_dport = htons (dport);

  tcp_phdr->ip_src_t.ip_src_int = srcip;
  tcp_phdr->ip_dst_t.ip_dst_int = dstip;

  net_ipv4_send_segment (pkt, ip_hdr, tcp_hdr, tcp_phdr);
}

static void
net_ipv4_pkt_init_headers (struct ip_hdr_t *ip_hdr)
{
  u_int32_t rlong;
  u_int16_t rshrt;

  rlong = rand ();
  rshrt = rlong & 0xffff;
  ip_hdr->ip_v = 4;
  ip_hdr->ip_hl = 5;
  ip_hdr->ip_tos = 0;
  ip_hdr->ip_len = htons (sizeof *ip_hdr + sizeof (struct tcp_hdr_t));
  ip_hdr->ip_id = rshrt;
  ip_hdr->ip_off = htons (0);
  ip_hdr->ip_ttl = 30;
  ip_hdr->ip_p = IPPROTO_TCP;
}

static void
net_tcp_pkt_init_headers (struct tcp_hdr_t *tcp_hdr,
                          struct tcp_phdr_t *tcp_phdr, u_int32_t tcp_flags)
{
  u_int32_t rlong;
  u_int16_t rshrt;

  rlong = rand ();
  rshrt = rlong & 0xffff;
  tcp_hdr->tcp_seq = rlong;
  tcp_hdr->tcp_ackseq = rlong;
  tcp_hdr->doff = sizeof *tcp_hdr / 4;
  tcp_hdr->tcp_flags = tcp_flags;
  tcp_hdr->tcp_window = rshrt;
  tcp_hdr->tcp_urgptr = rshrt;

  tcp_phdr->pad = 0;
  tcp_phdr->p_proto = IPPROTO_TCP;
  tcp_phdr->p_len = htons (sizeof *tcp_hdr);
}

void
net_ipv4_pkt_init (struct net_pkt_t *pkt, sock_t rsock, u_int32_t tcp_flags)
{
  int buflen;

  if (pkt == NULL)
    return;

  buflen = sizeof (struct ip_hdr_t) + sizeof (struct tcp_hdr_t);
  obuf_allocate (&pkt->pkt_buf, buflen);
  memset (pkt->pkt_buf.buf, 0, buflen);

  buflen = sizeof (struct tcp_phdr_t) + sizeof (struct tcp_hdr_t);
  obuf_allocate (&pkt->ppkt_buf, buflen);
  memset (pkt->ppkt_buf.buf, 0, buflen);

  pkt->ip_hdr = pkt->pkt_buf.buf;
  pkt->tcp_hdr = pkt->pkt_buf.buf + sizeof (struct ip_hdr_t);
  pkt->tcp_phdr = pkt->ppkt_buf.buf;
  pkt->rsock = rsock;

  net_ipv4_pkt_init_headers (pkt->ip_hdr);
  net_tcp_pkt_init_headers (pkt->tcp_hdr, pkt->tcp_phdr, tcp_flags);
}

void
net_ipv4_pkt_free (struct net_pkt_t *pkt)
{
  if (pkt == NULL)
    return;

  obuf_free (&pkt->pkt_buf);
  obuf_free (&pkt->ppkt_buf);
}

static int
net_get_localip_getifaddrs (char *if_name, u_int32_t *ip_addr)
{
  struct ifaddrs *ifa_head;
  int result;

  result = -1;
  if (getifaddrs (&ifa_head) == 0)
    {
      struct ifaddrs *ifa_cur;

      ifa_cur = ifa_head;
      for (ifa_cur = ifa_head; ifa_cur; ifa_cur = ifa_cur->ifa_next)
        {
          if (ifa_cur->ifa_name != NULL &&
              ifa_cur->ifa_addr != NULL)
            {
              if (strcmp (if_name, (char *) ifa_cur->ifa_name) != 0 ||
                  ifa_cur->ifa_addr->sa_family != AF_INET ||
                  !(ifa_cur->ifa_flags & IFF_UP))
                continue;
            
              memcpy (ip_addr, &(((struct sockaddr_in *) ifa_cur->ifa_addr)->sin_addr),
                      sizeof *ip_addr);
              result = 0;
              break;
            }
        }

      freeifaddrs (ifa_head);
    }

  return (result);
}

#ifdef LINUX
#define MAX_INTERFACES 16

static int
net_get_localip_ioctl (char *if_name, u_int32_t *ip_addr)
{
  struct ifreq ifr[MAX_INTERFACES];
  struct ifconf ifc;
  int i, if_count, sock;

  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock == -1)
    return (-1);

  memset (&ifc, 0, sizeof ifc);
  ifc.ifc_buf = (char *) &ifr;
  ifc.ifc_len = sizeof ifr;

  if (ioctl (sock, SIOCGIFCONF, &ifc) == -1)
    return (-1);

  if_count = ifc.ifc_len / sizeof (struct ifreq);
  for (i = 0; i < if_count; i++)
    {
      if (ioctl (sock, SIOCGIFADDR, &ifr[i]) == -1)
        continue;
      if (ioctl (sock, SIOCGIFFLAGS, &ifr[i]) == -1)
        continue;

      if (strcmp (if_name, (char *) ifr[i].ifr_name) != 0 ||
          !(ifr[i].ifr_flags & IFF_UP))
        continue;

      memcpy (ip_addr, &(((struct sockaddr_in *) &ifr[i].ifr_addr)->sin_addr),
              sizeof *ip_addr);
      break;
    }

  if (close (sock) != 0)
    return (-1);

  return (0);
}
#endif

int
net_get_localip (char *if_name, u_int32_t *ip_addr)
{
#ifdef LINUX
  if (net_get_localip_ioctl (if_name, ip_addr) != -1)
    return (0);
#endif

  if (net_get_localip_getifaddrs (if_name, ip_addr) != -1)
    return (0);

  return (-1);
}

void
net_iplistv4_print (struct iplistv4_t *list)
{
  while (list != NULL)
    {
      fprintf (stdout, "    %s/%d\n", nlookup (list->ip_t.ip_int),
               list->mask);
      list = list->next;
    }
}

static struct iplistv4_t *
net_iplistv4_add (struct iplistv4_t *head, u_int32_t ip, u_int32_t mask)
{
  struct iplistv4_t *ptr;

  if (mask <= 0 || mask > IPV4_MAX_MASK)
    return (NULL);

  ptr = malloc (sizeof (struct iplistv4_t));
  if (ptr == NULL)
    fatal ("%s: net_iplistv4_add: out of memory allocating %d-bytes\n",
           arg_progname, sizeof (struct iplistv4_t));

  ptr->next = head;
  ptr->ip_t.ip_int = ip;
  ptr->mask = mask;

  return (ptr);
}

void
net_iplistv4_free (struct iplistv4_t *head)
{
  while (head != NULL)
    {
      struct iplistv4_t *ptr;

      ptr = head->next;
      free (head);
      head = ptr;
    }
}

unsigned long long
net_iplistv4_sum (struct iplistv4_t *head)
{
  unsigned long long cur_sum;

  if (head == NULL)
    return (0);

  cur_sum = 0;
  while (head != NULL)
    {
      if (head->mask < IPV4_MAX_MASK)
        cur_sum += IPV4_MASK (head->mask);

      cur_sum++;
      head = head->next;
    }

  return (cur_sum);
}

struct iplistv4_t *
net_parse_iplistv4 (char *arg)
{
  struct iplistv4_t *head;
  char *bgn, *end;
  char host[HOSTNAME_MAX];

  /* empty the list */
  head = NULL;

  bgn = arg;
  do
    {
      struct iplistv4_t *cur_head;
      char *ptr;
      struct in_addr addr;
      u_int32_t mask;
      int len, result;

      end = strchr (bgn, ',');
      if (end)
        len = end - bgn;
      else
        len = strlen (bgn);

      if (len > sizeof host - 1)
        goto cleanup;

      strncpy (host, bgn, len); /* XXX - signedness bug */
      host[len] = '\0';         /* XXX - checked */

      if ((ptr = strchr (host, '/')))
        {
          *ptr++ = '\0';        /* remove '/' */
          if (strlen (ptr) != strspn (ptr, "0123456789"))
            goto cleanup;

          mask = atoi (ptr);
        }
      else
        mask = IPV4_MAX_MASK;

      result = inet_aton (host, (struct in_addr *) &addr);
      if (!result)
        goto cleanup;

      if ((cur_head = net_iplistv4_add (head, addr.s_addr, mask)) == NULL)
        goto cleanup;
      else
        head = cur_head;

      bgn = end + 1;
    }
  while (end != NULL);

  return (head);

cleanup:
  net_iplistv4_free (head);
  return (NULL);
}

struct iplistv4_t *
net_parse_iplistv4_file (char *filen, FILE * filen_fp)
{
  struct iplistv4_t *head;
  char host[HOSTNAME_MAX];

  /* empty the list */
  head = NULL;

  while (fgets (host, sizeof host, filen_fp))
    {
      struct iplistv4_t *cur_head;
      char *ptr;
      struct in_addr addr;
      u_int32_t mask;
      int result;

      /* chomp the newline */
      ptr = host + strlen (host) - 1;
      while (ptr >= host && (*ptr == '\r' || *ptr == '\n'))
        {
          *ptr = '\0';
          if (ptr > host)
            ptr--;
        }

      /* blank line */
      if (ptr == host)
        continue;

      if ((ptr = strchr (host, '/')))
        {
          *ptr++ = '\0';        /* remove '/' */
          if (strlen (ptr) != strspn (ptr, "0123456789"))
            goto cleanup;

          mask = atoi (ptr);
        }
      else
        mask = 32;

      result = inet_aton (host, (struct in_addr *) &addr);
      if (!result)
        goto cleanup;

      if ((cur_head = net_iplistv4_add (head, addr.s_addr, mask)) == NULL)
        goto cleanup;
      else
        head = cur_head;
    }

  return (head);

cleanup:
  net_iplistv4_free (head);
  return (NULL);
}

struct iplistv4_iter_t *
net_iplistv4_iter_init (struct iplistv4_iter_t *iter, struct iplistv4_t *list)
{
  u_int32_t net_addr;
  u_int32_t broad_addr;
  u_int32_t mask;

#ifdef DEBUG
  if (iter == NULL || list == NULL)
    return (NULL);
#endif

  iter->list = list;

  net_addr = htonl (list->ip_t.ip_int);
  broad_addr = htonl (list->ip_t.ip_int);

  if (list->mask < IPV4_MAX_MASK)
    {
      mask = IPV4_MASK (list->mask);
      net_addr &= ~mask;
      broad_addr |= mask;
    }

  iter->net_addr_t.ip_int = ntohl (net_addr);
  iter->broad_addr_t.ip_int = ntohl (broad_addr);
  iter->cur_ipv4_t.ip_int = iter->net_addr_t.ip_int;

  return (iter);
}

inline enum iter_t
net_iplistv4_iter_next (struct iplistv4_iter_t *iter, struct in_addr *out)
{
  /* XXX endianness - portability problems here */
  u_int32_t cur_n, broad_n;

#ifdef DEBUG
  if (iter == NULL)
    return (ITER_ERROR);
#endif

  cur_n = htonl (iter->cur_ipv4_t.ip_int);
  broad_n = htonl (iter->broad_addr_t.ip_int);
  if (cur_n < broad_n)
    {
      IPV4_INC (iter->cur_ipv4_t.ip_int);
      if (out)
        *out = iter->cur_ipv4_t.ip;

      return (ITER_CONTINUE);
    }
  else
    {
      if (iter->list->next)
        {
          net_iplistv4_iter_init (iter, iter->list->next);
          return (net_iplistv4_iter_next (iter, out));
        }
    }

  return (ITER_FINISH);
}

void
net_portlist_print (struct portlist_t *list)
{
  while (list != NULL)
    {
      fprintf (stdout, "    %d-%d\n", list->lower, list->upper);
      list = list->next;
    }
}

static struct portlist_t *
net_portlist_add (struct portlist_t *head, u_int32_t lower, u_int32_t upper)
{
  struct portlist_t *ptr;

  if (lower > upper)
    return (NULL);

  ptr = malloc (sizeof (struct portlist_t));
  if (ptr == NULL)
    fatal ("%s: net_portlist_add: out of memory allocating %d-bytes\n",
           arg_progname, sizeof (struct portlist_t));

  ptr->next = head;
  ptr->lower = lower;
  ptr->upper = upper;

  return (ptr);
}

void
net_portlist_free (struct portlist_t *head)
{
  while (head != NULL)
    {
      struct portlist_t *ptr;

      ptr = head->next;
      free (head);
      head = ptr;
    }
}

unsigned long long
net_portlist_sum (struct portlist_t *head)
{
  unsigned long long cur_sum;

  if (head == NULL)
    return (0);

  cur_sum = 0;
  while (head != NULL)
    {
      cur_sum += head->upper - head->lower + 1;
      head = head->next;
    }

  return (cur_sum);
}

struct portlist_iter_t *
net_portlist_iter_init (struct portlist_iter_t *iter, struct portlist_t *list)
{
#ifdef DEBUG
  if (iter == NULL || list == NULL)
    return (NULL);
#endif

  iter->list = list;
  iter->cur_port = list->lower;

  return (iter);
}

enum iter_t
net_portlist_iter_next (struct portlist_iter_t *iter, u_int16_t * out)
{
#ifdef DEBUG
  if (iter == NULL)
    return (ITER_ERROR);
#endif

  /* assert (iter->cur_port >= iter->list->lower); */
  if (iter->cur_port < iter->list->upper)
    {
      iter->cur_port++;
      if (out)
        *out = iter->cur_port;

      return (ITER_CONTINUE);
    }
  else
    {
      if (iter->list->next)
        {
          net_portlist_iter_init (iter, iter->list->next);
          return (net_portlist_iter_next (iter, out));
        }
    }

  return (ITER_FINISH);
}

struct portlist_t *
net_parse_portlist (char *arg)
{
  struct portlist_t *head;
  char *bgn, *end;
  char port[PORTLIST_MAX];

  /* empty the list */
  head = NULL;

  bgn = arg;
  do
    {
      struct portlist_t *cur_head;
      char *lptr, *uptr;
      u_int16_t lower, upper;
      int len;

      end = strchr (bgn, ',');
      if (end)
        len = end - bgn;
      else
        len = strlen (bgn);

      if (len > sizeof port - 1)
        goto cleanup;

      strncpy (port, bgn, len); /* XXX - signedness bug */
      port[len] = '\0';         /* XXX - checked */

      lptr = port;
      if ((uptr = strchr (lptr, '-')))
        {
          *uptr++ = '\0';       /* remove '-' */
          if (strlen (lptr) != strspn (lptr, "0123456789"))
            goto cleanup;

          if (strlen (uptr) != strspn (uptr, "0123456789"))
            goto cleanup;

          lower = atoi (lptr);
          upper = atoi (uptr);
        }
      else
        {
          if (strlen (lptr) != strspn (lptr, "0123456789"))
            goto cleanup;

          lower = upper = atoi (lptr);
        }

      if ((cur_head = net_portlist_add (head, lower, upper)) == NULL)
        goto cleanup;
      else
        head = cur_head;

      bgn = end + 1;
    }
  while (end != NULL);

  return (head);

cleanup:
  net_portlist_free (head);
  return (NULL);
}

struct portlist_t *
net_parse_portlist_file (char *filen, FILE * filen_fp)
{
  struct portlist_t *head;
  char port[PORTLIST_MAX];

  /* empty the list */
  head = NULL;

  while (fgets (port, sizeof port, filen_fp))
    {
      struct portlist_t *cur_head;
      char *ptr, *lptr, *uptr;
      u_int16_t lower, upper;

      /* chomp the newline */
      ptr = port + strlen (port) - 1;
      while (ptr >= port && (*ptr == '\r' || *ptr == '\n'))
        {
          *ptr = '\0';
          if (ptr > port)
            ptr--;
        }

      /* blank line */
      if (ptr == port)
        continue;

      lptr = port;
      if ((uptr = strchr (lptr, '-')))
        {
          *uptr++ = '\0';       /* remove '-' */
          if (strlen (lptr) != strspn (lptr, "0123456789"))
            goto cleanup;

          if (strlen (uptr) != strspn (uptr, "0123456789"))
            goto cleanup;

          lower = atoi (lptr);
          upper = atoi (uptr);
        }
      else
        {
          if (strlen (lptr) != strspn (lptr, "0123456789"))
            goto cleanup;

          lower = upper = atoi (lptr);
        }

      if ((cur_head = net_portlist_add (head, lower, upper)) == NULL)
        goto cleanup;
      else
        head = cur_head;
    }

  return (head);

cleanup:
  net_portlist_free (head);
  return (NULL);
}
